/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Roger Leszczynski <rogerles@gmail.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Test remote server's media stream
 *
 * \author Roger Leszczynski <rogerles@gmail.com>
 * 
 * \ingroup applications
 */

/*  This code is for the general public to use and modify at their
 *  will, as long as proper credits are given...any one who has
 *  worked on this code please append your name and email so we 
 *  know who to contact with issues 
 *  Copyright (c) 2006 Roger Leszczynski <rogerles@gmail.com>
 */

 
#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 42715 $")

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "asterisk.h"

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/module.h"
#include "asterisk/translate.h"
#include "asterisk/dsp.h"
#include "asterisk/utils.h"
#include "asterisk/options.h"
#include "asterisk/app.h"
#include "asterisk/cli.h"
#include "asterisk/pbx.h"

FILE *CHANCONF_ANA;

int ana_context_exists; /* Avoid rereading text file if extens exist */
int ana_channel; 	/* What channel are we sending to            */
char grab_user_in_analysis[256]; /* Current user we are looking at */

static char *app = "Analyze";

static char analyze_synopsis[]= "Analyze the media stream from a remote server call";

static char *synopsis = analyze_synopsis; 

static char analyze_descrip [] = 

"This application uses the cli command analyze to send a call out to a remote\n"
"target server with extension your specify, which calls back here on that\n"
"extension to analyze the packets of the media stream.\n\n"
"Usage 'Analyze <channel to dial>'\n"
"Ex: 'Analyze SIP/12345@targetmachine' or 'Analyze IAX2/user:pass@host/12345'\n\n"
"Make sure on the other end you are making a call back to here with the extension\n"
"being asterisk*app*analyze...\n"
"Ex:\n[whatever-your-incoming-context-for-this-machine-is]\n"
"exten => <exten-you-entered-in-this-command>,1,Dial(SIP/asterisk*app*analyze@<this-machine>)\n\n"
"Results written in asterisk log directory under file analysis.<user target>\n\n";

static char *descrip = analyze_descrip;

/*
 * <sob story>
 * Well ...I thought I was contributing something special when I 
 * wrote this, after compeletion of testing it I found out
 * app_test already exists... wtf....  ahh well the difference is
 * that it uses a coammnd in the dialplan while I do it throught
 * the cli </sob story>
 */



static int handle_analyze_channel(int fd, int argc, char *argv[])
{	
	
	int make_call_to_ana=0;
	int splitc=0;
	int srch_context_ana=0;  /* Lock check if extensions added	*/

	/* cmd line parse vars */
	char tech_to_ana[256];   /* Technoogy we are using	        */
	char dest_to_ana[256];   /* Call desination		        */
	char pdest_to_ana[256];  /* Copy of call dest for parsing again */
	char pd_confuser[256];   /* ConfUser 			        */
	char brkt_confuser[256]; /* [ConfUser] 			        */
	char read_dt_ana[256];   /* Copy of command argv	        */
	char *psplit,*psplit2;   /* Pointers to splits in cmd string    */
	int anacontext_exists = 0;  /* Is there context= in .conf file?	*/
	int anauser_exists = 0;  /* User found in .conf?		*/

	/* conf file search vars */
	char path_confana[PATH_MAX] = "/etc/asterisk"; /* config file	*/
	char *fd_confuser;	 /* [ConfUser]			        */
	char *fcontext_ana;    	 /* Search for context=		        */
	char fresult_ana[256] = "default"; /* Actual Context	        */
	char buffer_ana[512];    /* Buffer line of file we are reading  */
	char user_term_ch;	 /* User terminating char		*/
	char *fpsplit, *fsplitn; /* Pointers to splits in file line     */
	
	

	if (argc > 2 || argc == 1) 
		return RESULT_SHOWUSAGE;
	else{
		
		strcpy(read_dt_ana, argv[1]);
    		
		psplit = strchr(read_dt_ana,'/');
		
		if (psplit==NULL)
			return RESULT_SHOWUSAGE;
		else
			splitc=1;
    		
		*psplit = '\0';
    		
		strcpy(tech_to_ana,read_dt_ana);
    		strcpy(dest_to_ana, psplit+1);  
		
			
		if (splitc == 1){
		
			/*
			 * Channel structs vars from conf files are not shared,
 			 * so to make this app fully independant i have my own
			 * file parsers for vars i seek
			 */
		
			if (strcmp(tech_to_ana, "SIP") == 0) 	
				ana_channel = 1;

			if (strcmp(tech_to_ana, "IAX2") == 0) 	
				ana_channel = 2;
			
			if (ana_channel > 0) {
				
				strcpy(pdest_to_ana,dest_to_ana);
				psplit2 = strchr(pdest_to_ana, '@');
		
				if (psplit2 == NULL){
					ast_verbose("\nImproper Syntax...\n");
					return RESULT_SHOWUSAGE;
					
					}
		
				*psplit2 = '\0';
				strcpy(pd_confuser, psplit2+1);

				strcpy(path_confana, "\0");
				strcpy(path_confana, ast_config_AST_CONFIG_DIR);
				
				if (ana_channel == 1) {
					strcat(path_confana, "/sip.conf");
					CHANCONF_ANA = fopen(path_confana, "r");
					}

				if (ana_channel == 2) {
					strcat(path_confana, "/iax.conf");
					CHANCONF_ANA = fopen(path_confana, "r");
			        	}
				
				
				if (CHANCONF_ANA == NULL){  return 0;  }

				/* Search for [user] then context= */
				while (fgets(buffer_ana, 512, CHANCONF_ANA)){

					strcpy(brkt_confuser, "[");
					strcat(brkt_confuser, pd_confuser);
					strcat(brkt_confuser, "]");

					fd_confuser = strstr(buffer_ana, brkt_confuser);
	
					if (fd_confuser != NULL){
						ast_verbose("\nSIP USER %s found!\n", pd_confuser); 
						srch_context_ana = 1; 
						anauser_exists = 1;
					}
					
					if (srch_context_ana == 1 && ana_context_exists == 0) {
						fcontext_ana = strstr(buffer_ana, "context=");
						
						if (fcontext_ana != NULL) {

							anacontext_exists = 1;
							fpsplit = strchr(fcontext_ana, '=');
							*fpsplit = '\0';
							strcpy(fresult_ana, fpsplit+1); 
							
							if (ana_channel == 1)
								 user_term_ch = '\n';
							else
								 user_term_ch = '/';				
							
							fsplitn = strchr(fresult_ana, user_term_ch);
							*fsplitn = '\0'; /* I think that did it */
							
							srch_context_ana=0;
							
							if (fresult_ana != NULL) {	
								ast_context_create(NULL, fresult_ana, "app_analyze");
								ast_add_extension(fresult_ana , 1, "asterisk*app*analyze" /*ext*/,1 /*priority*/, NULL, NULL, "Analyze", "\0", NULL, "app_analyze.so");
								ast_add_extension(fresult_ana , 1, "asterisk*app*analyze" /*ext*/,2 /*priority*/, NULL, NULL, "Hangup", "\0", NULL, "app_analyze.so");
								ana_context_exists = 0;
							}
							else	
								return RESULT_SHOWUSAGE; 
						}
					}
				}
				fclose(CHANCONF_ANA);
			}
			
			if (anauser_exists == 1 && anacontext_exists == 1){
				ast_verbose("\nSending %s call to %s\n",tech_to_ana,dest_to_ana);
				
				make_call_to_ana = ast_pbx_outgoing_app(tech_to_ana, AST_FORMAT_SLINEAR, dest_to_ana,30000, "Playback", "1khz-ast", 0, 2, "0", "Analytical Test Call", NULL, 0, NULL);
				
				ana_channel=0;
				
				strcpy(grab_user_in_analysis, pd_confuser); /* pass as global var to analyze() */				
					
			}
		}

		else
			return RESULT_SHOWUSAGE;

	}

	if (make_call_to_ana < 0)

		return RESULT_SHOWUSAGE;

	return 0;

}

/* Here is were we analyze the call */
static int analyze_exec(struct ast_channel *chan, void *data)
{
	/* Similarities to app_record, vars changed incase of searching something for that app*/
	int res=0;
        unsigned int read_frame_data=0;
        struct ast_module_user *u;
        struct ast_frame *f = NULL;
        struct ast_dsp *sildet = NULL;  /* silence detector dsp          */
        int totalsil_count = 0;
        int dspsilence = 0;
        int silence    = 0;             /* 5 sec silence is long enough  */
        int sil_does_exist = 0;         /* silence timeout check         */
	int maxcallduration =0;
	int gottimeout =0;
	int chan_line_down = 0;
        int chan_line_noanswer = 0;
        int rfmt = 0;
        int waitres;

	FILE *analytical_file;		/* output results to ast log dir   */
	char pathlog_ana[PATH_MAX] = "/var/log/asterisk";  /* default path */

	float p_ana_good       = 0; 	/* Percentage of good packets */
	float p_ana_bad        = 0;	/* Percentage of bad packets  */
	float analyzed_good    = 0;	/* Good packet count 	      */
	float analyzed_bad     = 0;	/* Corrupted packet count     */
	float analyzed_total   = 0;	/* Total packet count         */
	float analyzed_unknown = 0;	/* Unknown packet count       */
	
	struct ast_silence_generator *silgen = NULL;

        u = ast_module_user_add(chan);

        if (chan->_state != AST_STATE_UP) {
                if (chan_line_down) {
                        /* At the user's option, skip if the line is not up */
                        ast_module_user_remove(u);
                        return 0;
                } else if (!chan_line_noanswer) {
                        /* Otherwise answer unless we're supposed to record while on-hook */
                        res = ast_answer(chan);
                }
        }

        if (res) {
                ast_log(LOG_WARNING, "Could not answer channel '%s'\n", chan->name);
                goto out;
        }

        if (ast_opt_transmit_silence)
                silgen = ast_channel_start_silence_generator(chan);

        if (maxcallduration <= 0)
                maxcallduration = -1;

        while ((waitres = ast_waitfor(chan, maxcallduration)) > -1) {
                if (maxcallduration > 0) {
                        if (waitres == 0) {
                                gottimeout = 1;
                                break;
                        }
                        maxcallduration = waitres;
                }

                f = ast_read(chan);
		analyzed_total++;

                if (!f) {
                        analyzed_bad++;
			res = -1;
                        break;
                }
		/* Only voice for now */
                if (f->frametype == AST_FRAME_VOICE) {

                        read_frame_data = ntohl(*((unsigned int *)(f->data)));  /* %08x */
                       /*ast_verbose("\n-> FRAME READ: %08x\n", read_frame_data   ); */
		       
		        if (read_frame_data==0xff908990 || read_frame_data==0x7f908990) 
				analyzed_good++;
			else
				analyzed_bad++;
	
                        if (silence > 0) {
                                dspsilence = 0;
                                ast_dsp_silence(sildet, f, &dspsilence);
                                if (dspsilence) {
                                        analyzed_bad++;
					totalsil_count = dspsilence;
                                } else {
                                        totalsil_count = 0;
					if (read_frame_data == 0xff908990 || read_frame_data == 0x7f908990) 
						analyzed_good++;
					else
						analyzed_bad++;
                                }
                                if (totalsil_count > silence) {
                                        /* Ended happily with silence */
                                        ast_frfree(f);
                                        sil_does_exist = 1;
                                        break;
                                }
                        }
                } else if (f->frametype != AST_FRAME_VOICE) {
			analyzed_bad++;
        	}
		ast_frfree(f);
	}
        if (!f) {
	        ast_log(LOG_DEBUG, "Got hangup\n");
                res = -1;
        }


        if (silgen)
                ast_channel_stop_silence_generator(chan, silgen);

 out:
        if ((silence > 0) && rfmt) {
                res = ast_set_read_format(chan, rfmt);
                if (res)
                        ast_log(LOG_WARNING, "Unable to restore read format on '%s'\n", chan->name);
                if (sildet)
                        ast_dsp_free(sildet);
        }

        ast_module_user_remove(u);

	p_ana_good = (analyzed_good / analyzed_total) * 100;
    	p_ana_bad = (analyzed_bad / analyzed_total) * 100;
    	analyzed_unknown = ((analyzed_total - (analyzed_good + analyzed_bad) ) / analyzed_total) * 100;
        
	ast_verbose("\n\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
        ast_verbose("%.2f percent of voice packets are going through properly.\n" , p_ana_good);
        ast_verbose("%.2f percent of voice packets are dead air.\n", p_ana_bad);
        ast_verbose("%.2f unknown packets\n", analyzed_unknown);
        ast_verbose("%0.f total packets.\n", analyzed_total);
        ast_verbose("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n\n");
	
	strcpy(pathlog_ana, "\0");
	strcpy(pathlog_ana, ast_config_AST_LOG_DIR);
	strcat(pathlog_ana, "/analysis.");

	strcat(pathlog_ana, grab_user_in_analysis);
	
	
	analytical_file = fopen(pathlog_ana, "w");

	if (analytical_file != NULL) { 
        	fprintf(analytical_file, "%.2f percent of voice packets are going through properly.\n" , p_ana_good);
        	fprintf(analytical_file, "%.2f percent of voice packets are dead air.\n", p_ana_bad);
        	fprintf(analytical_file, "%.2f unknown packets\n", analyzed_unknown);
        	fprintf(analytical_file, "%0.f total packets.\n", analyzed_total);
		ast_verbose("\nWriting Data to: %s\n\n", pathlog_ana);
	}

	fclose(analytical_file);


	return res;
}



static struct ast_cli_entry cli_analyze[] = {
	{ { "analyze", NULL },
	handle_analyze_channel, analyze_synopsis,
	analyze_descrip, NULL } };


static int unload_module(void)
{

	
	int res;

	res = ast_unregister_application(app);
        
	ast_cli_unregister_multiple(cli_analyze, sizeof(cli_analyze) / sizeof(struct ast_cli_entry));
	
	ast_module_user_hangup_all();

	return res;	
}

static int load_module(void)
{
	/* initilize vars*/
	ana_context_exists = 0;
	ana_channel = 0;

	strcpy(grab_user_in_analysis, "default");

	ast_cli_register_multiple(cli_analyze, sizeof(cli_analyze) / sizeof(struct ast_cli_entry));

	return ast_register_application(app, analyze_exec, synopsis, descrip);
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Media Stream Analyzer");
